# $Id: berkeleydb_storage.py 344 2007-05-03 12:19:09Z jcea $
#
# (c) 2006-2007 jcea@argo.es / jcea@hispasec.com
#
# PGP/GPG Public Key:
# pub   1024R/9AC53D4D 1995-01-01
#       Key fingerprint = F4 07 90 C2 58 86 8A 75  45 40 33 1C 72 4C E5 E1
# uid                  Jesus Cea Avion <jcea@hispasec.com>
# uid                  Jesus Cea Avion <jcea@argo.es>
#
# This product is covered by the GNU PUBLIC LICENSE, VERSION 2.
# For more detailt, read the file "LICENSE" in the distribution.

__version__="20070503"


# Try to use the newest
# BerkeleyDB bindings available
try :
  import bsddb
except :
  bsddb=None
try :
  import bsddb3
except :
  if not bsddb :
    raise RuntimeError("You need a BerkeleyDB instalation, \
                        and the corresponding python bindings")
  bsddb3=bsddb
if not bsddb :
  bsddb=bsddb3
if bsddb3.__version__.split(".") >= bsddb.__version__.split(".") :
  bsddb=bsddb3


from durus.storage import Storage
from durus.utils import str_to_int8,str_to_int4,int8_to_str,int4_to_str
from durus.serialize import unpack_record, split_oids
from durus.logger import log, is_logging

autoremove_logs=True
VERSION=int4_to_str(2)
OID_ROOT=int8_to_str(0)

class garbage_collection_interface(object) :
  bsddb_flags=0

  def __init__(self,dbenv,db,dbmeta,checkpoint,txn) :
    try :
      garbage=dbmeta.get("garbage",txn=txn,default=-1)
      if garbage==-1:
        garbage=""
        dbmeta.put("garbage","",txn=txn)

      count=dbmeta.get("count",txn=txn)
      if not count :
        count=int8_to_str(0)
        dbmeta.put("count",count,txn=txn)
      self._count=str_to_int8(count)

      if txn :
        txn.commit()
    except :
      if txn :
        txn.abort()

    self._garbage=set(split_oids(garbage))
    logging_debug = is_logging(20)
    logging_debug and log(20, '%d garbage objects', len(self._garbage))

    self._sync_delayed=set()

  def object_count(self) :
    return self._count

  def garbage_count(self) :
    return len(self._garbage)

  def could_read(self) :
    pass

  def clear_read(self) :
    pass

  def could_write(self) :
    pass

  def clear_write(self) :
    pass

  def update(self,count_update,a_borrar,txn=None) :
    if count_update :
      self._count+=count_update
      self._dbmeta.put("count",int8_to_str(self._count),txn=txn)

    l=len(self._garbage)
    a_borrar-=self._garbage
    if a_borrar :
      self._garbage.update(a_borrar)
      self._dbmeta.put("garbage","".join(a_borrar),dlen=0,doff=l*8,txn=txn)

    self._sync_delayed.update(a_borrar)

  def sync(self) :
    a=self._sync_delayed
    self._sync_delayed=set()
    return a

  def do(self,txn=None,must_do=False) :
    raise NotImplementedError

  def close(self) :
    raise NotImplementedError

class garbage_collection_noop(garbage_collection_interface) :
  bsddb_flags=0

  def do(self,txn=None,must_do=False) :
    pass

  def close(self) :
    pass

class garbage_collection_inline(garbage_collection_interface) :
  #bsddb_flags=bsddb.db.DB_THREAD

  def __init__(self,dbenv,db,dbmeta,checkpoint,txn) :
    self._dbenv=dbenv
    self._db=db
    self._dbmeta=dbmeta
    self._checkpoint=checkpoint
    garbage_collection_interface.__init__(self,dbenv,db,dbmeta,checkpoint,txn)

  def do(self,txn=None,must_do=False) :
    mas_borrados=set()
    txn2=txn
    if not txn :
      txn=self._dbenv.txn_begin(flags=bsddb.db.DB_TXN_NOSYNC) # Si falla, ya reintentaremos
    try :
      if self._garbage :
        l=len(self._garbage)-1
        #deleted_oid=self._garbage.pop()
        deleted_oid=self._dbmeta.get("garbage",dlen=8,doff=l*8,txn=txn) # Borramos del final de la lista
        self._garbage.remove(deleted_oid)
        obj=self._db.get(deleted_oid,txn=txn)
        if obj :
          self._dbmeta.delete(deleted_oid,txn=txn)
          self._db.delete(deleted_oid,txn=txn)
          self._count-=1
          self._dbmeta.put("count",int8_to_str(self._count),txn=txn)
          a,b,refs=unpack_record(obj)
          posibles_borrados=set(split_oids(refs))
          for i in posibles_borrados :
            count=self._dbmeta.get(i,txn=txn)
            assert count
            count=str_to_int4(count)-1
            if count :
              self._dbmeta.put(i,int4_to_str(count),txn=txn)
            else :
              mas_borrados.add(i)

# Si "mas_borrados" no tiene nada, vamos limpiando el registro
          self._garbage.update(mas_borrados)
          assert len(self._garbage)==(l+len(mas_borrados))
          self._dbmeta.put("garbage","".join(mas_borrados),dlen=8,doff=l*8,txn=txn) # Los borrados nuevos sobreescriben el anterior

      if not txn2 :
        txn.commit()
        txn2=True
        if must_do :
          # When doing a full and explicit "pack", and the
          # checkpoint is done in background, we could
          # get a lot of DB "log" files, since we could
          # generate transactions too fast for the cleanup
          # thread.
          self._dbenv.txn_checkpoint(16384,43) # 16 MB or 43 minutes
        else :
          self._checkpoint.do()
    except :
      if not txn2 :
        txn.abort()
      raise
    return not not self._garbage # Still work to do?

  def close(self) :
    del self._dbmeta
    del self._db
    del self._dbenv
    del self._checkpoint

  def __del__(self) :
    self.close()


class garbage_collection_inline_prefetch(garbage_collection_inline) :
  bsddb_flags=bsddb.db.DB_THREAD

  def __init__(self,dbenv,db,dbmeta,checkpoint,txn) :
    garbage_collection_inline.__init__(self,dbenv,db,dbmeta,checkpoint,txn)
    import threading
    self._prefetch_pending=False
    self._prefetch_done=True # Bootstrap the prefetch
    self._prefetch_notif=threading.Event()
    self._must_die=False
    t=threading.Thread(target=self._prefetch_thread,name="Garbage Collection Prefetch thread")
    t.setDaemon(True)
    t.start()
    self._thread=t
    self._prefetch_mutex=threading.Lock()

  def update(self,count_update,a_borrar,txn=None) :
    if count_update :
      self._count+=count_update
      self._dbmeta.put("count",int8_to_str(self._count),txn=txn)

    l=len(self._garbage)
    a_borrar-=self._garbage
    if a_borrar :
      self._garbage.update(a_borrar)
      if l :
        l-=1
        old_garbage=self._dbmeta.get("garbage",dlen=8,doff=l*8) # Add the old tail to the new tail
        self._dbmeta.put("garbage","".join(a_borrar)+old_garbage,dlen=8,doff=l*8,txn=txn)
      else :
        garbage="".join(a_borrar)
        self._dbmeta.put("garbage",garbage,txn=txn)
        garbage=garbage[-8:]
        self._prefetch_pending=garbage
        self._prefetch_done=False
        self._prefetch_notif.set()

    self._sync_delayed.update(a_borrar)

  def do(self,txn=None,must_do=False) :
    if not len(self._garbage) :
      return False

    if must_do :
      self._prefetch_mutex.acquire()
      try :
        more_to_do=garbage_collection_inline.do(self,txn,must_do)
      finally :
        self._prefetch_mutex.release()
      return not not more_to_do

    if not self._prefetch_done :
      return True # Still doing prefetch
    if self._prefetch_done!=self._prefetch_pending : # New work to do. We delete objects in a LIFO way
      self._prefetch_done=False
      l=len(self._garbage)-1
      deleted_oid=self._dbmeta.get("garbage",dlen=8,doff=l*8) # Borramos del final de la lista
      assert len(deleted_oid)==8
      self._prefetch_pending=deleted_oid
      self._prefetch_notif.set()
      return True # Doing a new prefetch

# We can't acquire the lock because if this function
# is called from "storage.end()", we will have
# a deadlock.
#
# So better be sure we call "garbage_collection_inline.do()"
# ONLY when the background prefetcher is idle.
# That is, when "self._prefetch_done==self._prefetch_pending".

    more_to_do=garbage_collection_inline.do(self,txn)
    self._prefetch_done=False

    if more_to_do :
      l=len(self._garbage)-1
      deleted_oid=self._dbmeta.get("garbage",dlen=8,doff=l*8) # Borramos del final de la lista
      assert len(deleted_oid)==8
      self._prefetch_pending=deleted_oid
      self._prefetch_notif.set()
      return True # Doing a new prefetch
    else :
      self._prefetch_pending=False
      return False

  def close(self) :
    self._must_die=True
    self._prefetch_notif.set()
    self._thread.join() # Be sure the thread is dead

    garbage_collection_inline.close(self)
    del self._thread

  def could_write(self) :
    self._prefetch_mutex.acquire()

  def clear_write(self) :
    self._prefetch_mutex.release()

  def _prefetch_thread(self) :
    try :
      while True :
        self._prefetch_notif.wait()
        oid=self._prefetch_pending
        self._prefetch_notif.clear()
        if self._must_die :
          return
        assert len(oid)==8
        self._prefetch_mutex.acquire()
        try :
          obj=self._db.get(oid)
          self._dbmeta.get(oid)
        finally :
          self._prefetch_mutex.release()
        if obj :
          a,b,refs=unpack_record(obj)
          refs=set(split_oids(refs))
          for i in refs :
            self._prefetch_mutex.acquire()
            try :
              self._dbmeta.get(i)
            finally :
              self._prefetch_mutex.release()
        self._prefetch_done=oid
    except :
      def do(*dummy,**dummy2) : # Propagate the exception to the main thread
        raise RuntimeError("El thread de 'Garbage Collection' ha muerto")
      self.do=do
      raise


class checkpoint_interface(object) :
  def __init__(self,dbenv) :
    """The dbenv parameter is a private (not shared) BerkeleyDB environment.

    This object is responsible of closing that DBEnv object when necessary.
    """

    raise NotImplementedError

  def do(self) :
    """This method is called when the storage instance thinks it is a good
    time to do a checkpoint. This object can do the actual checkpoint at
    its own, using these calls as hints.
    """

    raise NotImplementedError

  def close(self) :
    """Cleanup (in special, close the DBEnv if necessary).
    """

    raise NotImplementedError

class checkpoint_noop(checkpoint_interface) :
  def __init__(self,dbenv) :
    dbenv.close()

  def do(self) :
    pass

  def close(self) :
    pass

class checkpoint_inline(checkpoint_interface) :
  def __init__(self,dbenv) :
    self._dbenv=dbenv

  def do(self) :
    self._dbenv.txn_checkpoint(16384,43) # 16 MB or 43 minutes

  def close(self) :
    self._dbenv.close()
    del self._dbenv

  def __del__(self) :
    self.close()

class checkpoint_thread(checkpoint_interface) :
  def __init__(self,dbenv) :
    self._dbenv=dbenv
    self._must_die=False
    import threading
    self._event=threading.Event()
    t=threading.Thread(target=self._checkpoint_thread,name="Checkpointing thread")
    t.setDaemon(True)
    t.start()
    self._thread=t

  def do(self) :
    self._event.set()

  def close(self) :
    self._must_die=True
    self._event.set()
    self._thread.join() # Be sure the thread is dead

    self._dbenv.close()
    del self._dbenv
    del self._thread

  def _checkpoint_thread(self) :
    try :
      while True :
        self._event.wait()
        self._event.clear()
        if self._must_die :
          return
        self._dbenv.txn_checkpoint(16384,43) # 16 MB or 43 minutes
    except :
      def do(*dummy,**dummy2) : # Propagate the exception to the main thread
        raise RuntimeError("El thread de 'checkpointing' ha muerto")
      self.do=do
      raise

  def __del__(self) :
    self.close()
  
class checkpoint_thread_final_checkpoint(checkpoint_thread) :
  def close(self) :
    self._dbenv.txn_checkpoint() # Do a checkpoint if there is any DB change
    checkpoint_thread.close()

def monitor_read(func) :
  def f(*args,**kwargs) :
    args[0]._garbage_collection.could_read()
    try :
      return func(*args,**kwargs)
    finally :
      args[0]._garbage_collection.clear_read()
  return f

def monitor_write(func) :
  def f(*args,**kwargs) :
    args[0]._garbage_collection.could_write()
    try :
      return func(*args,**kwargs)
    finally :
      args[0]._garbage_collection.clear_write()
  return f
      
class BerkeleyDBStorage(Storage) :
  def __init__(self,directoryname,type="hash",cachesize=16*1024*1024,log_buffer_size=256*1024, \
               do_recover="",durable=True,async_log_write=True,read_only=False, \
               checkpoint_method="",garbage_collection_method="",validate_garbage=False) :

    self._read_only=read_only
    self._durable=durable

    self._sync=set()


# To avoid raising a spurious exception if we can't
# instantiate corrently the storage. For example,
# an already locked storage.
    self._garbage_collection=None
    self._checkpoint=None


    if do_recover=="" :
      do_recover=True
      if read_only :
        do_recover=False

    if checkpoint_method=="" :
      if read_only :
        checkpoint_method=checkpoint_noop
      else :
        checkpoint_method=checkpoint_thread

    if garbage_collection_method=="" :
      if read_only :
        garbage_collection_method=garbage_collection_noop
      else :
        garbage_collection_method=garbage_collection_inline_prefetch

    import os
    import fcntl
    assert directoryname
    self._lock=os.open(directoryname+os.sep+"storage.lock",os.O_RDWR | os.O_CREAT,0600)
    try :
      flags=fcntl.LOCK_EX
      if read_only :
        flags=fcntl.LOCK_SH
      fcntl.flock(self._lock,flags | fcntl.LOCK_NB)
    except :
       os.close(self._lock)
       raise RuntimeError("'%s' storage already locked" %directoryname)

    def config_dbenv() :
      dbenv=bsddb.db.DBEnv()
      dbenv.set_cachesize(0,cachesize)
      dbenv.set_lg_bsize(log_buffer_size) # Logging in-memory size
      return dbenv

    if type=="hash" :
      type=bsddb.db.DB_HASH
    elif type=="btree" :
      type=bsddb.db.DB_BTREE
    else :
      raise RuntimeError("Database type '%s' unsupported" %type)

# Don't use locking to avoid lock exhaustion when commiting
# a lot of objects.
    flags=bsddb.db.DB_INIT_TXN | bsddb.db.DB_INIT_MPOOL | bsddb.db.DB_INIT_LOG
    if not read_only :
      flags|=bsddb.db.DB_CREATE

    flags|=garbage_collection_method.bsddb_flags

    if not do_recover :
      try :
        self._dbenv=config_dbenv()
        self._dbenv.open(directoryname,flags)
        if self._dbenv.txn_stat()["nactive"] : # If transaction were in progress, do recover
          logging_debug = is_logging(20)
          logging_debug and log(20, 'Found an "in progress" transaction...')
          self._dbenv.close()
          do_recover=True
        else :
          f=0
          if read_only :
            f=bsddb.db.DB_RDONLY

          tmp=bsddb.db.DB(self._dbenv)
          try :
            tmp.open("storage.db",dbname="metadata",flags=f,dbtype=bsddb.db.DB_UNKNOWN)
            if tmp.get("in_use") :
              logging_debug = is_logging(20)
              logging_debug and log(20, 'The storage was not closed cleanly...')
              tmp.close()
              self._dbenv.close()
              do_recover=True
            del tmp
          except bsddb._db.DBNoSuchFileError :
            del tmp
      except bsddb._db.DBRunRecoveryError :
        self._dbenv.close()
        do_recover=True

    if do_recover :
      if read_only :
        raise RuntimeError("You can't open this storage in 'read only' mode now, because a 'db_recover' is required")
      self._dbenv=config_dbenv()
      logging_debug = is_logging(20)
      logging_debug and log(20, 'Doing a DB_RECOVER')
      self._dbenv.open(directoryname,flags | bsddb.db.DB_RECOVER)

    if autoremove_logs and not read_only :
      self._dbenv.set_flags(bsddb.db.DB_LOG_AUTOREMOVE,True)

    if read_only :
      txn=None
    else :
      txn=self._dbenv.txn_begin()

    try :
      try :
        tmp=bsddb.db.DB(self._dbenv)
        flags=0
        if read_only :
          flags=bsddb.db.DB_RDONLY
        tmp.open("storage.db",dbname="garbage_migration",flags=flags,dbtype=bsddb.db.DB_HASH,txn=txn)
        tmp.close()
        raise RuntimeError("Storage has an incomplete migration. You must delete it.")
      except bsddb._db.DBNoSuchFileError :
        pass

# We don't need to check "read only" mode in the
# storage methods because we already open the database
# files in "read only" mode if requested, so BerkeleyDB
# will catch automatically any write attempt.
      flags=0
      if read_only :
        flags=bsddb.db.DB_RDONLY

      flags|=garbage_collection_method.bsddb_flags

      self._db=bsddb.db.DB(self._dbenv)
      self._dbmeta=bsddb.db.DB(self._dbenv)
      try :
        self._db.open("storage.db",dbname="data",flags=flags,dbtype=bsddb.db.DB_UNKNOWN,txn=txn)
        self._dbmeta.open("storage.db",dbname="metadata",flags=flags,dbtype=bsddb.db.DB_UNKNOWN,txn=txn)
      except bsddb._db.DBNoSuchFileError :
        if not (flags & bsddb.db.DB_RDONLY) :
          flags|=bsddb.db.DB_CREATE
          self._db=bsddb.db.DB(self._dbenv)
          self._dbmeta=bsddb.db.DB(self._dbenv)
          self._db.open("storage.db",dbname="data",flags=flags,dbtype=type,txn=txn)
          self._dbmeta.open("storage.db",dbname="metadata",flags=flags,dbtype=type,txn=txn)
        else :
          raise

# You don't need to check the "read only" flag in the following statements because
# if the storage was not initialized, we already catched it.
# Remember that file creation and initialization is done in a
# single transaction, so if the process failed, no file were created, and
# then we already catched that.
#
# No metadata key CAN have a length of EXACTLY 8 chars, since then it
# could aliases a reference counter. That would be catastrofic. BEWARE!!
      version=self._dbmeta.get("version",txn=txn)
      do_count=False
      if version :
        if version==int4_to_str(0) :
          self._dbmeta.put("version",VERSION,txn=txn)
          do_count=True
          oid=self._dbmeta.get("oid",txn=txn)
          if not oid :
            self._dbmeta.put("oid",int8_to_str(str_to_int8(oid)+1),txn=txn)
        elif version==int4_to_str(1) :
          self._dbmeta.put("version",VERSION,txn=txn)
          oid=self._dbmeta.get("oid",txn=txn)
          if not oid :
            self._dbmeta.put("oid",int8_to_str(str_to_int8(oid)+1),txn=txn)
        else :
          assert version==VERSION
      else :
        self._dbmeta.put("version",VERSION,txn=txn)
        self.do_count=True

      if not read_only :
        self._dbmeta.put("in_use","*",txn=txn)

      self._oid=self._dbmeta.get("oid",txn=txn)
      if not self._oid :
        self._oid=OID_ROOT
        if not read_only :
          self._dbmeta.put("oid",self._oid,txn=txn) 
      self._oid=str_to_int8(self._oid)
      self._oid_changed=False

      if not read_only :
        import sys
        self._dbmeta.put(OID_ROOT,int4_to_str(sys.maxint/2),txn=txn) # ROOT NEVER DIE
        if do_count :
          self._dbmeta.put("count",int8_to_str(self._db.stat()["nkeys"]),txn=txn)
      if txn :
        txn.commit()
    except :
      if txn :
        txn.abort()
      raise
    self._txn_records={}

    if validate_garbage :
      object_exist=set()
      object_is_referenced=set([OID_ROOT]) # We don't want to delete the root object...
      cursor=self._db.cursor()
      record=cursor.first()
      while record :  # record=(key,data)
        oid,obj=record
        oid2,b,refs=unpack_record(obj)
        assert oid==oid2
        object_exist.add(oid)
        object_is_referenced.update(set(split_oids(refs)))
        record=cursor.next()
      garbage=object_exist-object_is_referenced
      del object_exist         # This object can be huge
      del object_is_referenced # This object can be huge
      logging_debug = is_logging(20)
      logging_debug and log(20, "validate_garbage: There are %d 'root' garbage objects", len(garbage))
      if not read_only :
        txn=self._dbenv.txn_begin()
        try :
          self._dbmeta.put("garbage","".join(garbage),txn=txn)
          txn.commit()
        except :
          txn.abort()
          raise

    if durable:
      self._begin_txn=lambda : self._dbenv.txn_begin()
    else :
      flags=bsddb.db.DB_TXN_NOSYNC
      if async_log_write :
        flags=bsddb.db.DB_TXN_WRITE_NOSYNC
      self._begin_txn=lambda : self._dbenv.txn_begin(flags=flags)

    dbenv=bsddb.db.DBEnv()
# We put the flags for BDB 4.3 compatibility
    dbenv.open(directoryname,bsddb.db.DB_INIT_TXN | bsddb.db.DB_INIT_MPOOL | bsddb.db.DB_INIT_LOG)
    self._checkpoint=checkpoint_method(dbenv)

    txn=None
    if not read_only :
      txn=self._dbenv.txn_begin(flags=bsddb.db.DB_TXN_NOSYNC)
    # The next call consumes the transaction
    self._garbage_collection=garbage_collection_method(self._dbenv,self._db,self._dbmeta,self._checkpoint,txn)

  def new_oid(self) :
    self._oid_changed=True
    oid=int8_to_str(self._oid)
    self._oid+=1
    return oid

  @monitor_read
  def load(self,oid) :
    self._garbage_collection.do()
    obj=self._db.get(oid)
    if not obj:
      obj={}[oid] # Raise an exception
    return obj

  def sync(self) :
    a=self._sync
    self._sync=set()
    a.update(self._garbage_collection.sync())
    return a

  def store(self, oid, record) :
    self._txn_records[oid]=record

  def begin(self) :
    self._txn_records={}

  @monitor_write
  def end(self, handle_invalidations=None) :
    counters={}
    borrados_diferidos=set()
    possible_resurrection=set()
    txn=self._begin_txn() # Can be durable or not, as configured
    try :
      if self._oid_changed :
        oid=int8_to_str(self._oid)
        self._dbmeta.put("oid",oid,txn=txn)

# Recalculate new reference counters
      for i,j in self._txn_records.iteritems() :
        if i not in counters : # No lo hemos procesado aun
          counters[i]=0

        obj=self._db.get(i,txn=txn)
        if obj :
          a,b,refs=unpack_record(obj)
          
          for k in split_oids(refs) :
            counters[k]=counters.get(k,0)-1
        else :
# Possible commit of a deleted object
          possible_resurrection.add(i)

        a,b,refs=unpack_record(j)
        for k in split_oids(refs) :
          counters[k]=counters.get(k,0)+1
        self._db.put(i,j,txn=txn)

      for i,j in counters.iteritems() :
        if not j : # Don't need to touch Reference-Counter
          if (i in possible_resurrection) and (i!=OID_ROOT) : # We rewrite a deleted object (but not ROOT)
            borrados_diferidos.add(i)
            self._dbmeta.put(i,int4_to_str(0),txn=txn) 
          continue
        count=self._dbmeta.get(i,txn=txn)
        if count :
          count=str_to_int4(count)+j
          assert count>=0
        else : # New object
          assert j>0
          count=j

        if count :
          self._dbmeta.put(i,int4_to_str(count),txn=txn)
        else :
          borrados_diferidos.add(i)

      self._garbage_collection.update(count_update=len(possible_resurrection),a_borrar=borrados_diferidos,txn=txn)
      self._garbage_collection.do(txn)
      txn.commit()
      txn=None
      self._checkpoint.do()
      self._oid_changed=False
    except :
      if txn :
        txn.abort()
      raise
    self._txn_records={}

    #if handle_invalidations and self._sync :
    #  a=self._sync
    #  self._sync=set()
    #  handle_invalidations(a)

  def get_packer(self) :
    def packer() :
      while self._garbage_collection.garbage_count() :
        yield self._garbage_collection.do(must_do=True)

      self._dbenv.txn_checkpoint() # A final checkpoint, when done
      #self._dbenv.txn_checkpoint(0,0,bsddb.db.DB_FORCE)
      #import os
      #for i in self._dbenv.log_archive(bsddb.db.DB_ARCH_ABS) :
      #  os.unlink(i)

    def packer_noop() :
      yield False

    if self._read_only :
      return packer_noop()
    else :
      return packer()

  def pack(self) :
    for i in self.get_packer() :
      pass

  def get_size(self) :
    return self._garbage_collection.object_count()

  def gen_oid_record(self, start_oid=None, batch_size=100):
    if start_oid==None :
      return self._gen_oid_record_berkeley_storage()
  # Call the generic code
    return Storage.gen_oid_record(start_oid,batch_size)

  def _gen_oid_record_berkeley_storage(self) :

# Don't do any writing while you are iterating, since
# you could skip or repeat records.
# See mailing list 20060417 - http://mail.mems-exchange.org/durusmail/durus-users/500/
# This method will give deleted objects not collected
# yet, just like the standard Durus storage.

    cursor=self._db.cursor()
    record=cursor.first()
    while record :  # record=(key,data)
      yield record
      record=cursor.next()

  def migrate(self,source) :
    assert self._dbmeta.get("oid")==OID_ROOT, "The destination storage must not exists previously"
    garbage=bsddb.db.DB(self._dbenv)
    txn=self._dbenv.txn_begin()
    try :
      garbage.open("storage.db",dbname="garbage_migration",flags=bsddb.db.DB_CREATE,dbtype=bsddb.db.DB_HASH,txn=txn) 
      txn.commit()
    except :
      txn.abort()
      raise

    count=0
    zero=int4_to_str(0)
    max_oid=OID_ROOT
    for oid,record in source :
      count+=1
      if max_oid<oid : # Can compare directly, since both or strings of eight bytes
        max_oid=oid
      txn=self._dbenv.txn_begin(flags=bsddb.db.DB_TXN_NOSYNC)
      oid2,obj,refs=unpack_record(record)
      assert oid==oid2
      self._db.put(oid,record,txn=txn)
      if not self._dbmeta.get(oid) :
        self._dbmeta.put(oid,zero,txn=txn)
        garbage.put(oid,"D",txn=txn)
      for i in split_oids(refs) :
        c=self._dbmeta.get(i,zero)
        self._dbmeta.put(i,int4_to_str(str_to_int4(c)+1),txn=txn)
        if garbage.get(i) :
          garbage.delete(i,txn=txn)

      txn.commit()
      self._dbenv.txn_checkpoint(16384,43) # 16 MB or 43 minutes

# The ROOT reference counter always exists,
# but to check it out never did any harm.
    assert not garbage.get(OID_ROOT) 

    gc=set()
    cursor=garbage.cursor()
    record=cursor.first()
    while record :  # record=(key,data)
      gc.add(record[0])
      record=cursor.next()

    txn=self._dbenv.txn_begin()
    self._dbmeta.put("garbage","".join(gc),txn=txn)
    self._dbmeta.put("oid",max_oid,txn=txn)
    self._dbmeta.put("count",int8_to_str(count))
    self._count=count
    txn.commit()
    self._dbenv.txn_checkpoint()

    garbage.close()
    bsddb.db.DB(self._dbenv).remove("storage.db",dbname="garbage_migration")
    self._dbenv.txn_checkpoint()

  def close(self) :
    if self._garbage_collection :
      self._garbage_collection.close()

    if not self._read_only :
      txn=self._dbenv.txn_begin() # Durable transaction
      self._dbmeta.delete("in_use",txn=txn)
      #if not self._durable :
      #  # We only do this to try to flush transaction log
      #  # to disk.
      #  self._dbmeta.put("count",int8_to_str(self._count),txn=txn)
      txn.commit()

    if self._checkpoint :
      self._checkpoint.close()

    self._dbmeta.close()
    self._db.close()
    self._dbenv.close()

    import os,fcntl
    fcntl.flock(self._lock,fcntl.LOCK_UN)
    os.close(self._lock)

  def __del__(self) :
    self.close()

