# $Id: monitor.py 36 2007-05-10 18:36:43Z jcea $

from __future__ import with_statement

# Esta clase es un constructor de decoradores
# utilizados para sincronizar el acceso al
# sistema de persistencia.
#
# Serian innecesarios si cada "thread" tuviese
# su conexion separada al sistema de persistencia,
# pero eso supondria tener una cache independiente,
# lo que consume memoria. Lo mas eficiente es tener
# una unica cache, compartida con todo el mundo.
# Pero en ese caso, hay que serializar los accesos.
#
# Cada monitor proporciona un acceso separado e
# independiente, asi que podemos tener lo mejor
# de ambos mundos.
#
# Esta clase, ademas, se encarga de gestionar los
# conflictos automaticamente, para no tener que
# preocuparnos de ello en el resto del programa.

class monitor(object) :
  def __init__(self,constructor) :
    import threading
    self.mutex=threading.Lock()
    self.conn=constructor()

  def __call__(self,func) :
    def _monitor(*args, **kwargs) :
      from durus.error import ConflictError
      while True : # Reintenta si hay conflictos
        with self.mutex :
          try : # Nos aseguramos de liberar el lock
            self.conn.abort() # Hacemos limpieza de cache
            ret=func(self.conn,*args, **kwargs)
            self.conn.commit()
            return ret
          except ConflictError :
            pass # El abort ya se hace en el bucle
          except :
            self.conn.abort()
            import sys
            import time
            print >>sys.stderr,time.ctime()
            raise

    return _monitor

  def get_conn(self) :
    return self.conn


