# $Id: servidor_web.py 471 2007-11-08 22:37:42Z alvaro $

from globales import monitor,allow_anonymous
from globales import http_max_clients
from globales import http_initial_timeout,http_keep_alive_timeout

urls={}

class ThreadingMixIn(object) :
  def __init__(self,num_threads=16) :
    import threading,Queue
    self._queue=Queue.Queue()
    for i in xrange(num_threads) :
      t=threading.Thread(target=self._thread)
      t.setDaemon(True)
      t.start()
      del t

  def _thread(self) :
    while True :
      request,client_addr=self._queue.get()
      try :
        self.finish_request(request, client_addr)
        self.close_request(request)
      except:
        self.handle_error(request, client_addr)
        self.close_request(request)

  def process_request(self,request,client_addr) :
    self._queue.put((request,client_addr))
    
def servidor_web() :
  from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler

  class handler(BaseHTTPRequestHandler) :
    if http_keep_alive_timeout : # Persistent connections
      protocol_version="HTTP/1.1"

    first_request=True

# The following method is ripped from BaseHTTPServer,
# implementing timeouts. When updating Python, coders
# should verify code consistence with BaseHTTPServer
# codebase.
    def handle_one_request(self):
        global http_initial_timeout,http_keep_alive_timeout
        import select
        v,dummy,dummy2=select.select([self.rfile],[],[],http_initial_timeout if self.first_request else http_keep_alive_timeout)
        self.first_request=False
        if v==[] :
          self.close_connection = 1
          return

        self.raw_requestline = self.rfile.readline()
        if not self.raw_requestline:
            self.close_connection = 1
            return
        if not self.parse_request(): # An error code has been sent, just exit
            return
        mname = 'do_' + self.command
        if not hasattr(self, mname):
            self.send_error(501, "Unsupported method (%r)" % self.command)
            return
        method = getattr(self, mname)
        method()

    def do_GET(self,post=False,head=False) :
      global urls,allow_anonymous
      cookie=self.headers.get("cookie",None)
      if cookie :
        import Cookie
        cookie=Cookie.SimpleCookie(cookie)
        @monitor
        def verifica(conn,cookie) :
          import database
          return database.usuario_verifica(conn,cookie=cookie)
        cpif_auth=cookie.get("cpif_auth")
        if cpif_auth :
          cookie=verifica(cpif_auth.value)
        else :
          cookie=None
      try :
        path=self.path
        q=path.find("?") # Aceptamos parametros "query" en la URL, necesarios para OpenID
        if q>=0 : path=path[:q]
        path=path.split("/")[1:]
        if not (cookie or allow_anonymous) :
          resultado=urls["LOGIN"](self,path,None)
        else :
          usuario,dummy=cookie or ("","")
          resultado=urls.get(path[0],None)
          if resultado :
            resultado=resultado(self,path,usuario)
          if not resultado : # URL inexistente o verificada como invalida
            resultado=urls["INVALIDA"](self,path,usuario)
      except :
        self.send_response(500)
        self.send_header("Content-Type","text/plain; charset=utf-8")
        self.end_headers()
        import traceback
        self.wfile.write(traceback.format_exc())
        import sys,time
        print >>sys.stderr,"EXCEPCION:",time.ctime()
        raise

      # We need this for HTTP persistent connections
      resultado[1]["Content-Length"]=len(resultado[2])

      self.send_response(resultado[0])

      for i,j in resultado[1].iteritems() :
        self.send_header(i,j)
      self.end_headers()

      if not head :
        self.wfile.write(resultado[2])

    def do_POST(self) :
      return self.do_GET(post=True)

    def do_HEAD(self) :
      return self.do_GET(head=True)

  class threadHTTPServer(ThreadingMixIn,HTTPServer) :
    def __init__(self,*args) :
       ThreadingMixIn.__init__(self,num_threads=http_max_clients)
       HTTPServer.__init__(self,*args)

  httpd=threadHTTPServer(("",8877),handler)
  httpd.serve_forever()

def calcula_urls() :
  global urls
  import os

  for i in os.listdir(os.path.curdir) :
    if i.startswith("url_") and i.endswith(".py") :
      modulo=i[:-len(".py")]
      nombre=modulo[len("url_"):]
      modulo=__import__(modulo)
      gestiona_url=getattr(modulo,"gestiona_url")
      assert nombre not in urls
      urls[nombre]=gestiona_url
      aliases=getattr(modulo,"url_aliases",[])
      for alias in aliases :
        assert alias not in urls
        urls[alias]=gestiona_url
      init=getattr(modulo,"init",None)
      if init : init()

calcula_urls()

